/**
 * Functions to take care of images with the help of libmagick, the
 * magickle image library. This reads images and converts them into
 * the proper format for the user interface to use.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <magick/magick.h>
#include <time.h>

#include "ui.h"
#include "image.h"

/**
 * Find the colour index corresponding to the given colours.
 * This does not try to find the closest match, but rather the exact
 * match, because the image is prepared to match the colourmap used.
 *
 * @param red The red component of the colour to find.
 * @param green The green component of the colour to find.
 * @param blue The blue component of the colour to find.
 *
 * @return the colour index matching the given colour, or a negative value
 * @return if the colour could not be found, which should not happen.
 */
#if 1
static inline int find_colour_index(int red, int green, int blue)
{
  int i, nr_of_colours;
  int diff, best_diff, best_match;
  int red_diff, green_diff, blue_diff;

  nr_of_colours = 1 << user_interface.ui_display.bit_depth;

  best_diff = 3 * 255 * 255;
  best_match = nr_of_colours - 1;
  for(i = 0 ; i < nr_of_colours ; i++) {
    red_diff   = red   - user_interface.ui_display.colourmap[3 * i + 0];
    green_diff = green - user_interface.ui_display.colourmap[3 * i + 1];
    blue_diff  = blue  - user_interface.ui_display.colourmap[3 * i + 2];
    /*
    diff = red_diff * red_diff + 
      green_diff * green_diff + 
      blue_diff * blue_diff;
      */
    diff = abs(red_diff) + abs(green_diff) + abs(blue_diff);
    if(diff == 0)
      return i;
    if(diff < best_diff) {
      best_diff = diff;
      best_match = i;
    }
  }

  return best_match;
}
#endif

/**
 *
 */
inline int get_real_colour(unsigned char *datap, unsigned char red, 
			   unsigned char green, unsigned char blue)
{
  u_int32_t colour;

  if(user_interface.ui_display.colourmap) {
    /* For palette based modes, we try to find the best matching
     * colour index.
     */
    datap[0] = find_colour_index(red, green, blue);
    return 1;
  } else {
    /* The display appears to be truecoloured. */
    colour = ((((red   >> (8 - user_interface.ui_display.red_length))   & 
		((1 << user_interface.ui_display.red_length)   - 1)) << 
	       user_interface.ui_display.red_offset) |
	      (((green >> (8 - user_interface.ui_display.green_length)) & 
		((1 << user_interface.ui_display.green_length) - 1)) << 
	       user_interface.ui_display.green_offset) |
	      (((blue  >> (8 - user_interface.ui_display.blue_length))  & 
		((1 << user_interface.ui_display.blue_length)  - 1)) << 
	       user_interface.ui_display.blue_offset));
    switch(user_interface.ui_display.bit_depth) {
    case 15:
    case 16:
      datap[0] = (colour >> 8) & 0xff;
      datap[1] = (colour >> 0) & 0xff;
      return 2;
      break;
    case 24:
      datap[0] = (colour >> 16) & 0xff;
      datap[1] = (colour >>  8) & 0xff;
      datap[2] = (colour >>  0) & 0xff;
      return 3;
      break;
    case 32:
      datap[0] = 0;
      datap[1] = (colour >> 16) & 0xff;
      datap[2] = (colour >>  8) & 0xff;
      datap[3] = (colour >>  0) & 0xff;
      return 4;
      break;
    default:
    }
  }

  return 0;
}

/**
 * Since libmagick, so far, is silly enough to runlength encode everything
 * we need to decode it here, to get the proper image data. This is said to
 * be changed in future versions of libmagick, starting from the upcoming
 * version 5.0. Then this will change.
 *
 * @param picture The image data where we are going to store the real image.
 * @param image A pointer to the image struct as provided by libmagick.
 */
static void runlength_decode(struct image_data *picture, Image *image)
{
  unsigned char *datap = picture->data;
  RunlengthPacket *rle_data = image->pixels;
  int x, y, i;
  unsigned char red, green, blue;

  i = 0;
  red = rle_data->red;
  green = rle_data->green;
  blue = rle_data->blue;
  for(y = 0 ; y < picture->height ; y++) {
    for(x = 0 ; x < picture->width ; x++) {
      i--;
      if(i < 0) {
	red = rle_data->red;
	green = rle_data->green;
	blue = rle_data->blue;
	i = rle_data->length;
	rle_data++;
      }
      datap += get_real_colour(datap, red, green, blue);      
    }
  }
}

#if 0
/**
 * This creates a colour cube containing the colours that the user interface
 * said it wanted. This is then used to map colours from all read images.
 * This is called once, just before the first image is read.
 *
 * Comments: When creating this, the MapImage() function rearranges the colours
 *           in the colourmap in what seems to be intensity order. This means
 *           the user interface must be aware of this as well.
 *           The problem is that I really want to be able to set the correct
 *           palette when opening the user interface. However, I certainly do
 *           not want a user interface to be dependent on libmagick.
 *           I have yet to find a good solution.
 *
 * @return a pointer to a libmagick Image struct with the correct colourmap.
 */
static Image *create_colour_cube(void)
{
  int i, nr_of_colours;
  Image *cube_image;
  ImageInfo cube_info;
  int red[256], green[256], blue[256];

  cube_image = AllocateImage(&cube_info);
  if(cube_image == NULL)
    return NULL;

  cube_image->pixels = (RunlengthPacket *)malloc(256 * sizeof(RunlengthPacket));
  if(cube_image->pixels == NULL)
    return NULL;
  cube_image->packets = 256;
  cube_image->columns = 256;
  cube_image->rows = 1;

  nr_of_colours = 1 << user_interface.ui_display.bit_depth;
  for(i = 0 ; i < nr_of_colours ; i++) {
    cube_image->pixels[i].red = 
      UpScale(user_interface.ui_display.colourmap[3 * i + 0]);
    cube_image->pixels[i].green = 
      UpScale(user_interface.ui_display.colourmap[3 * i + 1]);
    cube_image->pixels[i].blue = 
      UpScale(user_interface.ui_display.colourmap[3 * i + 2]);
    cube_image->pixels[i].length = 1;
  }

  MapImage(cube_image, cube_image, 0);

  /* Update the palette at the user interface. */
  for(i = 0 ; i < nr_of_colours ; i++) {
    user_interface.ui_display.colourmap[3 * i + 0] = 
      red[i] = cube_image->colormap[i].red;
    user_interface.ui_display.colourmap[3 * i + 1] = 
      green[i] = cube_image->colormap[i].green;
    user_interface.ui_display.colourmap[3 * i + 2] = 
      blue[i] = cube_image->colormap[i].blue;
  }
  ui_set_palette(red, green, blue);

  return cube_image;
}
#endif

/**
 * Reads an image from a stream. Earlier functions have decided that 
 * this is the proper function to use for this stream. Therefore, we
 * will not lay much weight on excessive error checking when it comes
 * to the possibility of wrong image types. 
 * The end product is converted into an appropriate format for the
 * user interface. 
 *
 * @param fd The file descriptor of the stream to read the image data from.
 * @param width The requested width of the image, or a negative value if the
 * @param width width is to be taken from the image.
 * @param height The requested height of the image, or a negative value if the
 * @param height height is to be taken from the image.
 *
 * @return an allocated struct with the converted image and information 
 * @return about it, or NULL if an error occurred.
 */
struct image_data *image_magick_read(int fd, int width, int height)
{
  FILE *fp;
  int new_fd;
  struct image_data *picture;
  Image *image, *scaled_image;
  ImageInfo image_info;
#if 0
  static Image *cube_image = NULL;
  time_t timer;
#endif

#if 0
  if(user_interface.ui_display.colourmap && cube_image == NULL) {
    cube_image = create_colour_cube();
    if(cube_image == NULL)
      return NULL;
  }
#endif

#if 0
  /* First duplicate the file descriptor, so that we can close it in the
   * correct way, without interfering with the method libmagick has chosen
   * to operate in.
   */
  new_fd = dup(fd);
  if(new_fd == -1) {
    /* If there is an error here, we could theoretically proceed with some
     * special treatment of the streams, but it indicates that something is
     * more wrong that we want it to be, so we just return an error here. 
     */
    return NULL;
  }

  fp = fdopen(new_fd, "r");
  if(fp == NULL) {
    close(new_fd);
    return NULL;
  }
#else
  {
    char buffer[16384];
    int bytes, left, written;

    new_fd = open("/tmp/zenimage", O_WRONLY | O_CREAT | O_TRUNC, 0644);
    if(new_fd < 0)
      return NULL;
    
    while((bytes = read(fd, buffer, 16384)) > 0) {
      left = bytes;
      while(left) {
	written = write(new_fd, buffer, left);
	left -= written;
	if(written < 0)
	  left = 0;
      }
    }
    close(new_fd);
    fp = NULL;
  }
#endif

  picture = (struct image_data *)malloc(sizeof(struct image_data));
  if(picture == NULL) {
    fclose(fp);
    return NULL;
  }

  /* Now it is finally time to let libmagick do its work. */
  GetImageInfo(&image_info);
  /*  image_info.file = fp; */
  strcpy(image_info.filename, "/tmp/zenimage");
  image_info.compression = NoCompression; 
  /*  image_info.colorspace = RGBColorspace; */
  image_info.colorspace = TransparentColorspace;
  /*  image_info.verbose = 1; */

  image = ReadImage(&image_info);
  if(image == NULL) {
    /* I do not know if ReadImage() closes the stream if an error occurs. 
     * Better safe than sorry.
     */
    fclose(fp);
    return NULL;
  }

  /* At this place, we should look at the width and height parameters to
   * this function, and if they differ from the width and height from the
   * recently read image, then we need to scale that image. Unless the given
   * width and height are negative, of course.
   */

  if(width < 0)
    picture->width = image->columns;
  else
    picture->width = width;
  if(height < 0)
    picture->height = image->rows;
  else
    picture->height = height;

  if(picture->width != image->columns || picture->height != image->rows) {
    scaled_image = ScaleImage(image, picture->width, picture->height);
    if(scaled_image) {
      DestroyImage(image);
      image = scaled_image;
    }
  }

#if 0
  /* Check with the user interface if we need to quantize the image. 
   * Right now, quantize everything when in palette based display mode. 
   * Later, check if the image is already palettized with the right amount
   * of colours in it.
   */
  if(user_interface.ui_display.colourmap && 
     (image->class == DirectClass || 
      image->colors > (1 << user_interface.ui_display.bit_depth))) {
    QuantizeInfo quantize_info;

    quantize_info.number_colors = 1 << user_interface.ui_display.bit_depth;
    quantize_info.tree_depth = user_interface.ui_display.bit_depth;
    quantize_info.dither = 0;
    quantize_info.colorspace = RGBColorspace;
    QuantizeImage(&quantize_info, image);
  }

  if(image->class == PseudoClass) {
    /* Map the colours of the image to the colour cube we have created for this 
     * purpose.
     */
    MapImage(image, cube_image, 1);
  }
#endif

  /* Allocate space for the image data in our struct. */
  picture->size = (picture->width * picture->height *
		   user_interface.ui_display.bit_depth) / 8;
  picture->data = (unsigned char *)malloc(picture->size);
  if(picture->data == NULL) {
    free(picture);
    DestroyImage(image);
    return NULL;
  }

#if 0
  timer = time(NULL);
  fprintf(stderr, "%d: %s", __LINE__, ctime(&timer));
#endif

  /* Decode the image, so that we can see what it is. */
  runlength_decode(picture, image);

#if 0
  timer = time(NULL);
  fprintf(stderr, "%d: %s", __LINE__, ctime(&timer));
#endif

  DestroyImage(image);

  return picture;
}
