/**
 * Functions used to do the actual laying out of all the parts that
 * were collected by the parser and put into the layout linked list.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "layout.h"
#include "ui.h"
#include "protocol.h"
#include "image.h"

/* External declaration of internal linked list for layout parts. 
 * The real variable is placed in build.c.
 */
extern struct layout_part *layout_internal_parts;

/* A pointer to the current page information for the active page. */
static struct layout_part *current_basep = NULL;

/**
 * If there are two or more parts after each other, that can be
 * put together, we do that here, in order to simplify future
 * processing of the parts in the layout process.
 * Currently, this is done for paragraphs and for pieces of text.
 *
 * Comments: It is possible that I will come to my senses later
 *           and integrate this into layout_add_part(). But for
 *           now, this works well, and looks nice enough to me.
 *
 * @param partp A pointer to the first element in a linked list of layout
 * @param partp parts. This is never NULL now, since we only call this
 * @param partp from layout_do(). At least so far. 
 *
 * @return non-zero value if an error occurred.
 */
static int layout_compress_parts(struct layout_part *partp)
{
  struct layout_part *tmp;

  while(partp) {
    switch(partp->type) {

    case LAYOUT_PART_PAGE_INFORMATION:
      layout_compress_parts(partp->child);
      break;
      
      /* If many paragraphs are placed after one another, html work in such 
       * a way that they act as if there was only one. The paragraph tag has
       * higher priority than the break row tag, which means if there is a
       * break row already, and a paragraph is approaching, the stored part
       * is saved as a paragraph.
       */
    case LAYOUT_PART_PARAGRAPH:
      if(partp->previous) {
	if(partp->previous->type == LAYOUT_PART_PARAGRAPH) {
	  if(!partp->previous->data.paragraph.paragraph) {
	    partp->previous->data.paragraph.paragraph = 
	      partp->data.paragraph.paragraph;
	  }
	  /* Remove the current part from the list, as it is now obsolete. */
	  if(partp->next != NULL)
	    partp->next->previous = partp->previous;
	  partp->previous->next = partp->next;
	  tmp = partp->previous;
	  layout_free_part(partp);
	  partp = tmp;
	} else if(partp->previous->previous && 
		  partp->previous->type == LAYOUT_PART_TEXT && 
		  partp->previous->previous->type == LAYOUT_PART_PARAGRAPH && 
		  (!strcmp(partp->previous->data.text.text, " ") || 
		   strlen(partp->previous->data.text.text) == 0)) {
	  if(!partp->previous->previous->data.paragraph.paragraph) {
	    partp->previous->previous->data.paragraph.paragraph = 
	      partp->data.paragraph.paragraph;
	  }

	  /* Remove both the almost empty text as well as the current paragraph. */
	  partp->next->previous = partp->previous->previous;
	  partp->previous->previous->next = partp->next;
	  tmp = partp->previous->previous;
	  layout_free_part(partp->previous);
	  layout_free_part(partp);
	  partp = tmp;
	} 
      }
      break;

    case LAYOUT_PART_LINK:
      layout_compress_parts(partp->child);
      break;

      /* This converts a graphics part to a text part if the user
       * interface has reported that it does not support images. 
       * I am not yet sure that this is the best place to do this. 
       * It might be better to let the parser do this directly when 
       * it accounters an img-tag. What is keeping me from doing so
       * is that I did not want to introduce the user interface with
       * the parser. I do not think they are ready to meet yet.
       */
    case LAYOUT_PART_GRAPHICS:
      if(!user_interface.ui_support.image) {
	if(partp->data.graphics.alt_text != NULL) {
	  struct layout_part *new_part;
	  char *new_text;
	  
	  new_part = layout_init_part(LAYOUT_PART_TEXT);
	  if(new_part == NULL) {
	    break;
	  }
	  new_part->align = partp->align;

	  new_text = (char *)malloc(strlen(partp->data.graphics.alt_text) + 1);
	  if(new_part == NULL) {
	    layout_free_part(new_part);
	  }
	  strcpy(new_text, partp->data.graphics.alt_text);
	  new_part->data.text.text = new_text;
	  if(partp->previous != NULL)
	    partp->previous->next = new_part;
	  if(partp->next != NULL)
	    partp->next->previous = new_part;
	  new_part->next = partp->next;
	  new_part->previous = partp->previous;
	  layout_free_part(partp);
	} else {
	  /* In this case the user interface does not support images, and 
	   * there is no alternative text. Then we simply remove the image
	   * from the parts list.
	   * 
	   * We are presented with an interesting problem. If the first part
	   * is an image with no alternative text, we get here, but cannot
	   * assign partp->previous->next a value, since partp->previous is 
	   * NULL. We also cannot remove the part, since we cannot update the 
	   * original from here. Of course it is quite easy to work around, 
	   * if we really want to. 
	   */
	  if(partp->previous != NULL) {
	    if(partp->next != NULL)
	      partp->next->previous = partp->previous;
	    partp->previous->next = partp->next;
	    layout_free_part(partp);
	  }
	}
      } else {
	/* Here, we find that the user interface does support images, and 
	 * therefore, we should read the image now.
	 */
	int img_fd;
	struct image_data *picture;

	/* If the image is already loaded for some reason, do not try to load 
	 * it again, or we will have problems.
	 */
	if(partp->data.graphics.data)
	  break;

	if(current_basep == NULL)
	  img_fd = protocol_open(partp->data.graphics.src, NULL, NULL);
	else
	  img_fd = protocol_open(partp->data.graphics.src, 
				 current_basep->data.page_information.url,
				 current_basep->data.page_information.base_url);

	/* If we are unable to open the image, we might want to convert the
	 * image part into a text part containing the alternative text.
	 */
	if(img_fd < 0)
	  break;
	
	if(partp->geometry.width == 0)
	  partp->geometry.width = -1;
	if(partp->geometry.height == 0)
	  partp->geometry.height = -1;

	picture = image_open(img_fd, partp->geometry.width, partp->geometry.height);
	if(picture == NULL) {
	  protocol_close(img_fd);
	  break;
	}

	partp->data.graphics.data = picture->data;
	partp->data.graphics.size = picture->size;
	partp->data.graphics.type = LAYOUT_PART_GRAPHICS_RAW;
	partp->data.graphics.alt_text = NULL;
	partp->geometry.width = picture->width;
	partp->geometry.height = picture->height;
	free(picture);

	protocol_close(img_fd);
      }
      break;

    default:
    }
    
    partp = partp->next;
  }
     
  return 0;
}

/**
 * Set width and height for all parts where it can be set at this point. 
 * Also keep track of the maximum width, because we need to return that
 * to the caller of this eminent function.
 *
 * @param partp A pointer to the first element in a linked list of layout
 * @param partp parts. This is never NULL now, since we only call this
 * @param partp from layout_do(). At least so far. 
 *
 * @return the single maximum width of a part, not counting a text as one
 * @return wide part, since that can be divided later.
 */
static int layout_set_sizes(struct layout_part *partp)
{
  int max_width, part_max;

  max_width = 0;
  part_max = 0;
  while(partp) {
    switch(partp->type) {
    case LAYOUT_PART_PAGE_INFORMATION:
      part_max = layout_set_sizes(partp->child);
      break;

    case LAYOUT_PART_PARAGRAPH:
      partp->geometry.width = 0;
      if(partp->data.paragraph.paragraph)
	partp->geometry.height = 8;
      else
	partp->geometry.height = 0;
      break;

    case LAYOUT_PART_TEXT:
      partp->geometry.width = 0;
      partp->geometry.height = partp->data.text.style.size;
      break;

    case LAYOUT_PART_LINK:
      part_max = layout_set_sizes(partp->child);
      break;

    case LAYOUT_PART_LINE:
      if(!partp->data.line.absolute)
	partp->data.line.width = (partp->data.line.width * 
				  user_interface.ui_display.width) / 100;
      partp->geometry.width = partp->data.line.width;
      partp->geometry.height = 2 * 7 + partp->data.line.size;
      break;

      /* Find out the width and height of this image, somehow. */
    case LAYOUT_PART_GRAPHICS:
      if(partp->geometry.width == 0)
	partp->geometry.width = 16;
      if(partp->geometry.height == 0)
	partp->geometry.height = 16;
      break;

    default:
      partp->geometry.width = 0;
      partp->geometry.height = 0;
    }

    if(part_max == 0) {
      if(partp->geometry.width > max_width)
	max_width = partp->geometry.width;
    } else {
      if(part_max > max_width)
	max_width = part_max;
    }

    partp = partp->next;
  }

  return max_width;
}

/**
 * This puts the correct alignment on previously layouted parts. That
 * means that it travels backwards in the linked list, and possibly moves
 * each part that are on the same row to a more correct position in 
 * the known universe. 
 *
 * @param partp A pointer to the last element in the linked list to be
 * @param partp aligned. If this is NULL, we exit gracefully, but also
 * @param partp returns a fabulous error code.
 * @param max_width The total width of the whole screen as we know it.
 * @param max_part_height The height of the highest part in this row.
 *
 * @return non-zero value if an error occurred.
 */
int layout_align_parts(struct layout_part *partp, int max_width, 
		       int max_part_height)
{
  int current_y_position;
  int row_width, row_x_offset;

  /* Skip all deleted parts. */
  while(partp && partp->type == LAYOUT_PART_DELETED)
    partp = partp->previous;

  if(partp == NULL)
    return 1;

  if(partp->type == LAYOUT_PART_PARAGRAPH)
    return 0;

  /* We read the Y position from the pointed part, and store that. This
   * is then used to see that we keep our alignment within the same row
   * of parts. We stop aligning as soon as we hit the previous row, since
   * that is already aligned.
   */
  current_y_position = partp->geometry.y_position;

  /* The row width is used to know how much we should move the parts
   * horizontally, if we align to the right or to the center.
   */
  row_width = partp->geometry.x_position + partp->geometry.width;

  switch(partp->align.horizontal) {
  case LAYOUT_PART_ALIGN_LEFT:
  case LAYOUT_PART_ALIGN_JUSTIFY:
    row_x_offset = 0;
    break;

  case LAYOUT_PART_ALIGN_CENTER:
  case LAYOUT_PART_ALIGN_FORCED_CENTER:
    row_x_offset = (max_width - row_width) / 2;
    break;
    
  case LAYOUT_PART_ALIGN_RIGHT:
    row_x_offset = max_width - row_width;
    break;

  default:
    row_x_offset = 0;
  }

  /* Step backwards through the list and align, until we bump into
   * the start of the list, or until the row comes to a beginning.
   * Also break if we get to a title or a paragraph.
   */
  while(partp) {
    if(partp->geometry.y_position < current_y_position ||
       partp->type == LAYOUT_PART_PAGE_INFORMATION ||
       partp->type == LAYOUT_PART_PARAGRAPH)
      break;

    partp->geometry.x_position += row_x_offset;

    /* Vertical alignment does not yet work properly. */
#if 0
    /* Align this particular part vertically. */
    switch(partp->align.vertical) {
    case LAYOUT_PART_ALIGN_TOP:
      /* We do not need to do any alignment here. */
      break;

    case LAYOUT_PART_ALIGN_MIDDLE:
      partp->geometry.y_position += (max_part_height - partp->geometry.height) / 2;
      break;

    case LAYOUT_PART_ALIGN_BOTTOM:
      partp->geometry.y_position += max_part_height - partp->geometry.height;
      break;

    default:
    }
#endif

    partp = partp->previous;
  }

  return 0;
}

/**
 * Perform the layouting of all the wonderful parts that the parser
 * has picked up from the fantastic stream provided by earlier parts
 * of the procedure. This will go through the linked list of layout
 * parts, and decide how they are to be spread over the screen that
 * the interface presents us with. This is done using extensive
 * mathematical calculations, optimised to their maximum.
 *
 * @param parts A pointer to the first element in a linked list of layout
 * @param parts parts. If `parts' is NULL, use the internal linked list.
 * @param keep_position A non-zero value means that the X and Y position
 * @param keep_position we are currently working at, is not changed from 
 * @param keep_position the previous call. This is used for layouting child
 * @param keep_position part trees. A value of zero will reset them to 0;0.
 *
 * @return non-zero value if an error occurred.
 */
int layout_do(struct layout_part *parts, int keep_position)
{
  struct layout_part *partp;
  static int max_width = 0, max_row_height = 0;
  static int x_position = 0, y_position = 0;

  if(parts == NULL)
    partp = layout_get_page_information(NULL);
  else
    partp = parts;

  /* If the internal list is completely empty, there is not much good we
   * can do, else than consider it an error and bring these devestating
   * news back to the unfortunate caller of this function.
   */
  if(partp == NULL)
    return 1;

  if(partp->type == LAYOUT_PART_PAGE_INFORMATION)
    current_basep = partp;

  if(!keep_position) {
    x_position = 0;
    y_position = 0;
    max_row_height = 0;

    /* Get rid of excess fat. Once. */
    layout_compress_parts(partp);

    /* Set width and height of all parts. This also find the maximum width 
     * among all parts. 
     */
    max_width = layout_set_sizes(partp);
    
    /* We want to use max_width as reference when layouting the parts. */
    if(max_width < user_interface.ui_display.width)
      max_width = user_interface.ui_display.width;
  }


  /* Here is an attempt to explain what this is all about.
   * We go through the linked list from beginning to end, working with one
   * part at a time, until that it completely finished.
   * The variables x_position and y_position are set to the coordinates
   * where the next part will end up. An exception to this rule is when we
   * stumble across a paragraph part, which instead sets those variables
   * to where the next part will be placed. Another exception is if we are
   * at the end of a line. Then we have to put the current part on the next
   * row, and set the position variables accordingly. In the case that the
   * part is a text, we might have to split that text up into several parts
   * if it is too long to fit on one row.
   * The variable max_row_height is used to keep track of the highest height
   * during one row. It is used when we go down to a new row, to know how much
   * to increase y_position with.
   * 
   * Alignments are handled in such a way that a row is first layouted as if
   * there was no alignment, or rather, as if the alignment was always to the
   * left and bottom, and parts are then moved accordingly to the real 
   * alignment.
   *
   * I hope this is all clear now.
   */

  /* Traverse the tree and lay out to the future. */
  while(partp) {
    switch(partp->type) {
    case LAYOUT_PART_PAGE_INFORMATION:
      layout_do(partp->child, 1);
      /* When we get here, we know the whole page is layouted. */
      return 0;

    case LAYOUT_PART_PARAGRAPH:
      x_position = 0;
      y_position += max_row_height;
      max_row_height = 0;
      partp->geometry.x_position = x_position;
      partp->geometry.y_position = y_position;
      y_position += partp->geometry.height;
      break;

    case LAYOUT_PART_LINE:
      x_position = 0;
      y_position += max_row_height;
      max_row_height = 0;
      partp->geometry.x_position = x_position;
      partp->geometry.y_position = y_position;
      y_position += partp->geometry.height;
      break;

    case LAYOUT_PART_TEXT:
      {
	int text_width, text_length;
	char *text;

	/* If the user interface does not support font fiddling, we set the
	 * font size to 8 pixels here.
	 */
	if(!user_interface.ui_support.font)
	  partp->data.text.style.size = 8;

	if(partp->data.text.style.size > max_row_height)
	  max_row_height = partp->data.text.style.size;
      
	/* Here we really need to ask the user interface how long a 
	 * specific string will be in pixels. There is no way we can
	 * know that here otherwise.
	 * For now, assume, like we have done so many times before.
	 */

	text = partp->data.text.text;
	/* We do not want any excess whitespace at the beginning of a row. 
	 * We also know that there can be only one whitespace at this point.
	 */
	if(x_position == 0 && isspace(text[0])) {
	  memmove(partp->data.text.text, partp->data.text.text + 1,
		  strlen(partp->data.text.text));
	}

	text_length = strlen(text);
	text_width = text_length * 8;

	if(x_position + text_width < max_width) {
	  /* If we manage to drag the program counter here, it means that all
	   * text in this part actually fits within the room that is left of
	   * the user interface. Despite the long comment, it is a good thing.
	   */
	  partp->geometry.x_position = x_position;
	  partp->geometry.y_position = y_position;
	  partp->geometry.width = text_width;
	  x_position += text_width;
	} else {
	  /* If we reach this point, we need to break the text up into two
	   * parts. The first part will of course fit in this row, or we would
	   * not have broken it in that place. The second part is put into a
	   * new layout part, which is discreetly sneaked into the linked list
	   * right after the current, and will therefore become the next part
	   * to be processed by this very same loop. Bejoyed.
	   *
	   * Currently, we cut the text right off, and does not care about 
	   * word wrapping. This will be fixed later.
	   */
	  int first_text_width, first_text_length, second_text_length;
	  struct layout_part *second_part;
	  char *second_text;

	  /* Place this part in the correct position. */
	  partp->geometry.x_position = x_position;
	  partp->geometry.y_position = y_position;
	  
	  /* Put the rest of the string into a new part. */

	  first_text_width = max_width - x_position;
	  first_text_length = first_text_width / 8;

	  /* Set the position variables with new values. */
	  x_position = 0;
	  y_position += max_row_height;
	  max_row_height = partp->data.text.style.size;

	  /* Find the end of a full word, where we can safely break the row. */
	  while(first_text_length > 0 && 
		!isspace(partp->data.text.text[first_text_length - 1]))
	    first_text_length--;

	  /* If the whole text happened to end up on the next line, just 
	   * put the current text there. In order to not having to write too
	   * much extra code here, we will simply copy the whole string to 
	   * the second text and remove this one completely. 
	   * To simplify for us, we mark the part as deleted by changing
	   * its type. It will then later be totally ignored by the renderer.
	   */
	  if(first_text_length == 0) {
	    partp->type = LAYOUT_PART_DELETED;
	  }

	  /* Now it is possible to know the length of this text part. */
	  partp->geometry.width = first_text_length * 8;

	  second_text_length = text_length - first_text_length;

	  second_text = (char *)malloc(second_text_length + 1);
	  if(second_text == NULL) {
	    /* Flee the field, before it is too late not to. */
	    break;
	  }
	  strcpy(second_text, text + first_text_length);

	  /* We cannot set the zero to end the first text until after we
	   * have copied the text to the second text. That is here. 
	   */
	  text[first_text_length] = '\0';

	  second_part = layout_init_part(LAYOUT_PART_TEXT);
	  if(second_part == NULL) {
	    /* Oops, that was not supposed to happen. Let us bail out. */
	    free(second_text);
	    break; 
	  }
	  second_part->data.text.text = second_text;
	  second_part->data.text.style = partp->data.text.style;
	  second_part->align = partp->align;	  

	  /* Add new part as the next part. */
	  second_part->next = partp->next;
	  second_part->previous = partp;
	  partp->next = second_part;
	}

	break;
      }

    case LAYOUT_PART_LINK:
      layout_do(partp->child, 1);
      break;

    case LAYOUT_PART_GRAPHICS:
      if(x_position + partp->geometry.width >= max_width) {
	/* Change row. */
	x_position = 0;
	y_position += max_row_height;
	max_row_height = 0;
      }
      partp->geometry.x_position = x_position;
      partp->geometry.y_position = y_position;
      x_position += partp->geometry.width;
      if(partp->geometry.height > max_row_height)
	max_row_height = partp->geometry.height;
      break;

    default:
    }

    if(partp->type == LAYOUT_PART_LINE) {
      layout_align_parts(partp, max_width, max_row_height);
    } else if(x_position == 0) {
      if(partp->previous) {
	if(partp->previous->type != LAYOUT_PART_LINE)
	  layout_align_parts(partp->previous, max_width, max_row_height);
      } else {
	layout_align_parts(partp->previous, max_width, max_row_height);
      }
    }

    if(partp->next == NULL) {
      /* Do not forget to align the last parts, please. */
      layout_align_parts(partp, max_width, max_row_height);
    }

    partp = partp->next;
  }

  return 0;
}
