/** 
 * Structs and functions prototypes that are used to define the
 * structure created by the parser and used by the layouter.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

#ifndef _LAYOUTER_LAYOUT_H_
#define _LAYOUTER_LAYOUT_H_

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>

/**
 * A generically used rectangle definition for generical use.
 *
 * @member x_position The X position of the upper left corner.
 * @member y_position The Y position of the upper left corner.
 * @member width The width of the rectangle.
 * @member height The height of the rectangle.
 */
struct layout_rectangle {
  int x_position;
  int y_position;
  int width;
  int height;
};

/**
 * The different types of parts there are.
 */
enum layout_part_type {
  LAYOUT_PART_DELETED = -1,
  LAYOUT_PART_PAGE_INFORMATION,
  LAYOUT_PART_PARAGRAPH,
  LAYOUT_PART_TEXT,
  LAYOUT_PART_LINK,
  LAYOUT_PART_GRAPHICS,
  LAYOUT_PART_LINE,
  LAYOUT_PART_TABLE,
  LAYOUT_PART_FORM
};

/**
 * The different alignments in the horizontal view of the world.
 */
enum layout_part_horizontal_align {
  LAYOUT_PART_ALIGN_LEFT,
  LAYOUT_PART_ALIGN_CENTER,
  LAYOUT_PART_ALIGN_RIGHT,
  LAYOUT_PART_ALIGN_JUSTIFY,
  LAYOUT_PART_ALIGN_FORCED_CENTER
};

/**
 * The different alignments as a vertical being.
 */
enum layout_part_vertical_align {
  LAYOUT_PART_ALIGN_TOP,
  LAYOUT_PART_ALIGN_MIDDLE,
  LAYOUT_PART_ALIGN_BOTTOM
};

/**
 * The different types of graphical data that are recognised.
 *
 * Comments: This might disappear as the program evolves. I believe
 *           the RAW type will be the only one actually used, making 
 *           this enumeration obsolete.
 */
enum layout_part_graphics_type {
  LAYOUT_PART_GRAPHICS_UNKNOWN,
  LAYOUT_PART_GRAPHICS_RAW,
  LAYOUT_PART_GRAPHICS_JPEG,
  LAYOUT_PART_GRAPHICS_PNG,
  LAYOUT_PART_GRAPHICS_GIF,
  LAYOUT_PART_GRAPHICS_TIFF
};

/**
 * Contains the title text for the page. There can be only one.
 *
 * @member text A pointer to the title text.
 * @member url The URL for the page represented by this page information.
 * @member base_url The base URL used as reference for all other objects
 * @member base_url which uses a relative URL, or NULL to use `url' as base.
 * @member background_colour The background colour of the page.
 * @member text_colour The text colour of the page.
 * @member link_colour The link colour of the page.
 * @member active_link_colour The active link colour of the page.
 * @member visited_link_colour The visited link colour of the page.
 */
struct layout_part_page_information_data {
  char *title;
  char *url;
  char *base_url;
  u_int32_t background_colour;
  u_int32_t text_colour;
  u_int32_t link_colour;
  u_int32_t active_link_colour;
  u_int32_t visited_link_colour;
};

/**
 * Contains information used to describe a paragraph. This includes
 * paragraphs, headers, row breaks et cetera.
 *
 * @member paragraph A non-zero value means the break is a paragraph
 * @member paragraph and a new line will be printed on its existance.
 */
struct layout_part_paragraph_data {
  int paragraph;
};

/**
 * Contains information to describe the current style of a text.
 *
 * @member size Size of the font to be used for this text.
 * @member colour The colour to render the text in.
 * @member monospaced A non-zero value means the text is rendered monospaced.
 * @member italic A non-zero value means the text is rendered italic.
 * @member bold A non-zero value means the text is rendered bold.
 * @member underlined A non-zero value means the text is rendered underlined.
 * @member overlined A non-zero value means the text is rendered overlined.
 * @member subscript A non-zero value means the text is rendered as subscript.
 * @member superscript A non-zero value means the text is rendered as 
 * @member superscript superscript.
 * @member preformatted A non-zero value means the text is rendered as
 * @member preformatted preformatted. This implies using a monospaced font.
 */
struct layout_text_styles {
  int size;
  u_int32_t colour;
  int monospaced;
  int italic;
  int bold;
  int underlined;
  int overlined;
  int subscript;
  int superscript;
  int preformatted;
};

/**
 * Contains the information needed in order to store a text, which is
 * the text itself accompanied by all its possible attributes.
 *
 * @member text A null terminated char array of the text.
 * @member style The style to describe this particular piece of text.
 */
struct layout_part_text_data {
  char *text;
  struct layout_text_styles style;
};

/**
 * Contains the information needed for a link.
 *
 * @member href A URL to where the link points, as written in the struct.
 */
struct layout_part_link_data {
  char *href;
};

/**
 * Contains the information needed in order to store a graphical object.
 * This is normally used for the img tag, but may also be used by internal
 * graphical objects.
 *
 * @member data A pointer to the graphical data for the object.
 * @member alt_text A null terminated char array containing the alternative
 * @member alt_text textual description of the graphical object, or NULL if
 * @member alt_text there is none.
 * @member src The URL to where to find the image.
 * @member size The size of the data in bytes.
 * @member border The thickness of the border. Zero means no border.
 * @member border_colour A 24-bit value describing the colour of the border.
 * @member border_colour This is used when the image is a link.
 * @member type The type of the data stored in `data'.
 */
struct layout_part_graphics_data {
  unsigned char *data;
  char *alt_text;
  char *src;
  size_t size;
  int border;
  u_int32_t border_colour;
  enum layout_part_graphics_type type;
};

/**
 * Contains information about a line, usually a horizontal one.
 *
 * @member width The width of the line.
 * @member absolute A non-zero value means the width is described as absolute 
 * @member absolute pixels, otherwise percentage of the width of the screen.
 * @member size The height of the line.
 */
struct layout_part_line_data {
  int width;
  int absolute;
  int size;
};

/**
 * Contains the data used by the different part types.
 *
 * @member generic Raw data used when no extra information is needed.
 * @member paragraph Used by paragraph breaks.
 * @member text Used by text parts.
 * @member graphics Used by graphical parts.
 */
union layout_part_data {
  void *generic;
  struct layout_part_page_information_data page_information;
  struct layout_part_paragraph_data paragraph;
  struct layout_part_text_data text;
  struct layout_part_link_data link;
  struct layout_part_graphics_data graphics;
  struct layout_part_line_data line;
};

/**
 * Contains information about how a part is to be aligned compared to the
 * surrounding parts. 
 *
 * @member horizontal How should the part should be aligned horizontally.
 * @member vertical How should the part should be aligned vertically.
 */
struct layout_aligns {
  enum layout_part_horizontal_align horizontal;
  enum layout_part_vertical_align vertical;  
};

/**
 * Defines a part of the screen that the layouter will
 * build from all parts.
 *
 * @member type The type of the part.
 * @member geometry The position and size of the part.
 * @member align How should the part be aligned.
 * @member data Optional data for some part types.
 * @member next Pointer to the next part placed next to this.
 * @member previous Pointer to the previous part placed before this.
 * @member child Pointer to the first part within this.
 */
struct layout_part {
  enum layout_part_type type;
  struct layout_rectangle geometry;
  struct layout_aligns align;
  union layout_part_data data;
  struct layout_part *next;
  struct layout_part *previous;
  struct layout_part *child;
};

/* Prototypes of layout build functions. */
extern struct layout_part *layout_init_part(enum layout_part_type type);
extern int layout_free_part(struct layout_part *part);
extern int layout_add_part(struct layout_part *new);
extern int layout_delete_all_parts(struct layout_part *parts);
extern struct layout_part *layout_get_page_information(struct 
						       layout_part *parts);
extern struct layout_part *layout_get_internal_parts(void);
extern struct layout_part *layout_init_page(char *url, char *referer);

/* Prototypes of layout layouting functions. */
extern int layout_do(struct layout_part *parts, int keep_position);

#endif /* _LAYOUTER_LAYOUT_H_ */
