/**
 * This is where it all begins.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>

#ifdef HAVE_GETOPT_H
#include <getopt.h>
#endif /* HAVE_GETOPT_H */

#include "settings.h"
#include "parse.h"
#include "ui.h"
#include "layout.h"
#include "protocol.h"

/* Prototypes of functions further on in this file. */
static void set_default_settings(void);
static int interpret_options(int argc, char *argv[]);
static void print_usage(char *program_name);
static void print_version(void);

/* The real structure. */
struct zen_settings settings;

/**
 * The beginning of the long and difficult journey.
 *
 * @param argc The number of arguments given to the program.
 * @param argv An array of strings with the given arguments.
 *
 * @return The program exit code, 0 if it exited normally.
 */
int main(int argc, char *argv[])
{
  int arg, ret;
  int fd;
  char *url;

  /* Initialize the settings. */
  set_default_settings();

  /* Find out if the user tries to tell us something. */
  arg = interpret_options(argc, argv);

  /* Remainings of an ancient time. */
  /* printf("Hello, world!\n"); */

  /* If there was not URL given on the command line, we brutally
   * force the user to the Zen pages.
   */
  if(arg < argc) {
    url = argv[arg];
  } else {
    url = "http://www.nocrew.org/software/zen/";
  }

  /* Setup the user interface of Our choice. 
   * We must initialize the user interface before request a document, 
   * to be able to send the correct user agent information to the server.
   */
  ret = ui_init();
  if(ret)
    return ret;

  /* Dump as source, means really dump it. So let us do just that. */
  if(settings.dump_source) {
    char *buffer;
    int bytes;

    /* Open a stream to the specified URL. */
    fd = protocol_open(url, NULL, NULL);
    if(fd < 0) {
      fprintf(stderr, "%s: Could not open %s\n", argv[0], url);
      ui_exit();
      return fd;
    }

    buffer = (char *)malloc(16384);
    if(buffer == NULL) {
      fprintf(stderr, "%s: Memory exhausted.\n", argv[0]);
      return 1;
    }

    while((bytes = read(fd, buffer, 16384)) > 0) {
      write(STDOUT_FILENO, buffer, bytes);
    }

    free(buffer);

    return 0;
  }

  /* Give over control to the user interface. */
  ret = ui_open(url);

  /* It looks here as if the program does not do anything, but it does.
   * The user interface has entered its event loop state, and is running
   * the show from now on. The only contact with the outer world, the
   * world into which I am writing this in, is through functions provided
   * for the user interface.
   */

  /* Later, the user interface will have the possibility to thread out into
   * its own thread using POSIX thread library. Then, this thread, the main
   * program will be suspended, waiting for calls to any of the helper
   * functions it has provided for the user interface. When the user interface
   * had had enough fun, it will tell the main program to end its activities,
   * and the whole program will end.
   */

  /* End all user interface activity. */
  ui_exit();

  return ret;
}

/**
 * Sets the default settings for this particular program, called Zen.
 * Later, this will read the settings from a configfile.
 */
void set_default_settings(void)
{
  settings.dump_source = 0;
  settings.interface = NULL;
}

/**
 * Interpret the options given to the program.
 *
 * @param argc The number of arguments given to the program.
 * @param argv An array of strings with the given arguments.
 *
 * @return The index of the first argument which did not contain an option.
 */
static int interpret_options(int argc, char *argv[])
{
  int arg;

#ifdef HAVE_GETOPT_LONG
  struct option long_options[] = {
    { "interface", required_argument, NULL, 'i' },
    { "source", no_argument, NULL, 's' },
    { "help", no_argument, NULL, 'h' },
    { "version", no_argument, NULL, 'V' },
    { NULL, 0, NULL, 0 } };
#endif /* HAVE_GETOPT_LONG */

  char *short_options = "i:shV";

  arg = 0;
  while(arg != EOF) {

#ifdef HAVE_GETOPT_LONG
    arg = getopt_long(argc, argv, short_options, long_options, NULL);
#else /* !HAVE_GETOPT_LONG */
    arg = getopt(argc, argv, short_options);
#endif /* !HAVE_GETOPT_LONG */

    switch(arg) {
    case EOF: /* End of arguments */
      break;

    case 'i': /* --interface */
      settings.interface = optarg;
      break;

    case 's': /* --source */
      settings.dump_source = 1;
      break;

    case 'h': /* --help */
      print_usage(argv[0]);
      exit(0);

    case 'V': /* --version */
      print_version();
      exit(0);

    default:
      fprintf(stderr, "-%c: Unknown option skipped\n", arg);
    }
  }
 
  return optind;
}

/**
 * Print the usage help text when the option --help is given.
 *
 * @param program_name The name of the program.
 */
static void print_usage(char *program_name)
{
  fprintf(stdout, 
	  "Usage: %s [OPTION(S)] [URL]\n"
	  "\n"
	  "Valid options are:\n"
#ifdef HAVE_GETOPT_LONG
	  "-i name        Select which user interface to use for this session\n"
	  "    --interface=name\n"
	  "-s  --source   Dump the object pointed to by URL on stdout\n"
	  "-h  --help     Print this text and exit\n"
	  "-V  --version  Print version and exit\n"
#else /* !HAVE_GETOPT_LONG */
	  "-i name        Select which user interface to use for this session\n"
	  "-s             Dump the object pointed to by URL on stdout\n"
	  "-h             Print this text and exit\n"
	  "-V             Print version and exit\n"
#endif /* !HAVE_GETOPT_LONG */
	  "\n"
	  "Available interfaces:\n"
	  "    %s\n"
	  , program_name, AVAILABLE_UI);
}

/**
 * Print version information when the option --version is given.
 */
static void print_version(void)
{
  fprintf(stdout, "%s version %s\n"
	  "\n"
	  "Available interfaces:\n"
	  "    %s\n"
	  , PACKAGE, VERSION, AVAILABLE_UI);
}
