/** 
 * Various functions that will aid the parsing of the input stream.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "tags.h"
#include "layout.h"
#include "states.h"
#include "helpers.h"

/**
 * As the name suggests, this function will get one byte from the stream.
 * It does this in a bit more efficient way than just doing a read(), since
 * it reads blocks of data to a buffer first.
 *
 * Comments: Right now, the number of bytes read each time is fixed. If need
 *           be, I will put this as a config file option in the future.
 *
 * @param fd The file descriptor from where to read data. 
 * @param c A pointer to a character where the byte is placed.
 *
 * @return code similar to what a read would return, that is a 1 if the byte
 * @return was read successfully, a 0 if the stream ended before we could read
 * @return the character, or a negative value if an error occurred. 
 */
int parse_get_byte(int fd, char *c)
{
  static char buffer[16384];
  static int index = 0, left = 0;
  
  if(left == 0) {
    left = read(fd, buffer, 16384);
    if(left <= 0)
      return left;
    index = 0;
  }

  *c = buffer[index++];
  left--;

  return 1;
}

/**
 * Allocates a parse_param struct and initialises its name and value.
 *
 * @param name The name of the parameter.
 * @param value The value of the parameter.
 *
 * @return an allocated parse_param struct or NULL if an error occurred.
 */
struct parse_param *parse_alloc_param(const char *name, const char *value)
{
  struct parse_param *paramp;
  int index;

  paramp = (struct parse_param *)malloc(sizeof(struct parse_param));
  if(paramp == NULL)
    return NULL;

  /* Initialise name */
  paramp->name = (char *)malloc(strlen(name) + 1);
  if(paramp->name == NULL)
    return NULL;

  /* Copy parameter name and convert it to lower case. */
  for(index = 0 ; index < strlen(name) ; index++)
    paramp->name[index] = tolower(name[index]);
  paramp->name[strlen(name)] = '\0';

  /* Initialise value */
  if(value == NULL) {
    /* This was a parameter without a value. */
    paramp->value = NULL;
  } else {
    /* Parameter with a value. */
    paramp->value = (char *)malloc(strlen(value) + 1);
    if(paramp->value == NULL)
      return NULL;

    strcpy(paramp->value, value);
  }

  paramp->next = NULL;

  return paramp;
}

/**
 * Free the complete memory of a parse_param struct.
 * It only frees the given object, and does not follow the
 * link to the next parameter.
 *
 * @param paramp A pointer to the parse_param object to be freed.
 *
 * @return a non-zero value if an error occurred.
 */
int parse_free_param(struct parse_param *paramp)
{
 if(paramp->name != NULL)
    free(paramp->name);

 if(paramp->value != NULL)
    free(paramp->value);

  return 0;
}

/**
 * Look very carefully for a parameter that matches the given parameter
 * name, and come up with the suitable value for that parameter.
 *
 * @member paramp A pointer to the first parse_param object in a linked list.
 * @member name The name of the parameter to look for.
 *
 * @return a pointer to the value for the given parameter name, or NULL
 * @return if the parameter name could not be found.
 */
char *parse_get_param_value(struct parse_param *paramp, char *name)
{
  while(paramp) {
    if(!strcmp(paramp->name, name))
      return paramp->value;

    paramp = paramp->next;
  }

  return NULL;
}

/**
 * Allocates the a parse_tag struct and initialises its name and type. 
 *
 * @param name The name of the parameter.
 * @param type The type of the parameter.
 *
 * @return an allocated parse_tag struct or NULL if an error occurred.
 */
struct parse_tag *parse_alloc_tag(const char *name, enum parse_tag_type type)
{
  struct parse_tag *tagp;
  int index;

  tagp = (struct parse_tag *)malloc(sizeof(struct parse_tag));
  if(tagp == NULL)
    return NULL;

  /* Initialise name */
  tagp->name = (char *)malloc(strlen(name) + 1);
  if(tagp->name == NULL)
    return NULL;

  /* Copy tag name and convert it to lower case. */
  for(index = 0 ; index < strlen(name) ; index++)
    tagp->name[index] = tolower(name[index]);
  tagp->name[strlen(name)] = '\0';

  /* Initialise type */
  tagp->type = type;

  /* Initialise linked list */
  tagp->first = NULL;

  return tagp;
}

/**
 * Free the complete memory of a parse_tag struct. 
 * Automatically frees the parameter list as well.
 *
 * @param tagp A pointer to the parse_tag object to be freed.
 *
 * @return a non-zero value if an error occurred.
 */
int parse_free_tag(struct parse_tag *tagp)
{
  struct parse_param *paramp;

  if(tagp->name != NULL)
    free(tagp->name);

  /* Free all allocated parameters from the list. */
  paramp = tagp->first;
  while(paramp != NULL) {
    struct parse_param *next_paramp = paramp->next;
    int retval;

    retval = parse_free_param(paramp);
    if(retval != 0)
      return retval;

    paramp = next_paramp;
  }

  return 0;
}

/**
 * Read one word from the input stream, up to a delimiter character.
 * Delimiter characters are defined in the string given by the
 * parameter `delimeters'.
 *
 * @param fd File descriptor of input stream.
 * @param word Pointer to a string variable where the word is stored.
 * @param first Put this character in as first character in the word.
 * @param first If it is 0 (zero), do not put any character in.
 * @param delimiters A string of characters which mark the end of the word.
 *
 * @return -1 if an error occurred, zero if the stream ended or
 * @return a positive number containing the delimiter character which 
 * @return ended the word.
 */
int parse_read_word(int fd, char **word, int first, char *delimiters)
{
  char *tmp=NULL, c;
  int current=0, index=0, done=0;
  int amount_read;

  while(!done) {
    /* Current buffer has come to an end, please allocate a new. */
    if(index == current) {
      char *newtmp;
      current += 1024;
      newtmp = realloc(tmp, current);
      if(newtmp == NULL) {
	if(tmp != NULL)
	  free(tmp);
	return -1;
      }
      tmp = newtmp;
    }
 
    if(first == 0) {
      /* Read one character from the stream. */
      amount_read = parse_get_byte(fd, &c);
    } else {
      /* Use the input character instead, and simulate a read. */
      c = first;
      amount_read = 1;
      first = 0;
    }

    if(amount_read <= 0) {
      /* If amount_read is zero, the stream ended. Interpret this as 
       * an error, just as if read returned -1.
       */
      free(tmp);
      return -1;
    } else if(strchr(delimiters, c) != NULL) {
      tmp[index++] = '\0';
      *word = (char *)realloc(tmp, index);
      if(*word == NULL) {
	free(tmp);
	return -1;
      }
      done = 1;
    } else {
      /* A character which is not a delimiter was read. */
      tmp[index++] = c;
    }	     
  }

  return (int)c;
}

/**
 * Skip characters in the input stream.
 * Which characters to skip are defined in the string given by the
 * parameter `skippers'. Since the last character should not be 
 * skipped, and we read through an unknown stream, we cannot put it
 * back into the stream. This function instead returns the first
 * character which was not skipped.
 *
 * @param fd File descriptor of input stream.
 * @param skippers A string of characters which are to be smoked.
 *
 * @return -1 if an error occurred, zero if the stream ended or
 * @return a positive number containing the first character which 
 * @return was not a skipper.
 */
int parse_skip_leading(int fd, char *skippers)
{
  char c;
  int done=0;
  ssize_t amount_read;

  while(!done) {
    /* Read one character from the stream. */
    amount_read = parse_get_byte(fd, &c);

    if(amount_read == -1) {
      /* An error occurred while reading. */
      return -1;
    } else if(amount_read == 0) {
      /* The stream ended in a mysterious way. */
      return 0;
    } else if(strchr(skippers, c) == NULL) {
      /* The character was not a skipper! */
      done = 1;
    }
  }
  return (int)c;
}

/**
 * Extract the tag and its parameters from the input stream.
 *
 * @param fd File descriptor of input stream.
 *
 * @return pointer to a parse_tag struct where the result is stored
 * @return or NULL if an error occurred.
 */
struct parse_tag *parse_get_tag(int fd)
{
  struct parse_tag *tagp;
  char *name;
  int c, ending;
  enum parse_tag_type type;

  c = parse_skip_leading(fd, " \t\n\r");
  if(c <= 0) {
    return NULL;
  }

  if(c == '/') { /* End tag */
    ending = parse_read_word(fd, &name, 0, "> \t\n\r");
    type = PARSE_TAG_END;
  } else { /* Start tag */
    ending = parse_read_word(fd, &name, c, "> \t\n\r");
    type = PARSE_TAG_START;
  }
  /* If we bumped into the end of the stream, we cannot consider this 
   * to be a valid tag
   */
  if(ending <= 0) {
    return NULL;
  }

  tagp = parse_alloc_tag(name, type);
  free(name);
  if(tagp == NULL) {
    return NULL;
  }

  /* We have already reaached the end of the tag. No parameters. */
  if(ending == '>')
    return tagp;

  /* Here we know a space character is the last one read. */
  ending = parse_skip_leading(fd, " \t\n\r");
  if(ending <= 0)
    return NULL;

  while(ending != '>') {
    char *param_name, *param_value;
    struct parse_param *tmpparamp;

    tmpparamp = NULL;

    /* Here we have the start of the parameters to the tag. */
    ending = parse_read_word(fd, &param_name, ending, "=> \t\n\r");
    if(ending <= 0)
      return NULL;

    if(ending != '>' && ending != '=') {
      ending = parse_skip_leading(fd, " \t\n\r");
      if(ending <= 0)
	return NULL;
    }

    if(ending != '=') {
      /* This parameter did not have a value. */
      tmpparamp = parse_alloc_param(param_name, NULL);
      free(param_name);
      if(tmpparamp == NULL)
	return NULL;

    } else if(ending == '=') {
      /* There is definitely a value to this parameter, or should be. */
      ending = parse_skip_leading(fd, " \t\n\r");
      if(ending <= 0) {
	free(param_name);
	return NULL;
      }

      /* If the first character is a quote, the parameter value should
       * only be terminated by a second quote.
       */
      if(ending == '"') {
	ending = parse_read_word(fd, &param_value, 0, "\"");
      } else {
	ending = parse_read_word(fd, &param_value, ending, "> \t\n\r");
      }
      if(ending <= 0) {
	free(param_name);
	return NULL;
      }

      if(ending != '>') {
	ending = parse_skip_leading(fd, " \t\n\r");
	if(ending <= 0) {
	  free(param_name);
	  free(param_value);
	  return NULL;
	}
      }
      
      /* Allocate the parameter struct. */
      tmpparamp = parse_alloc_param(param_name, param_value);
      free(param_name);
      free(param_value);
      if(tmpparamp == NULL)
	return NULL;
    }

    /* Put the new parameter, if any, first in the linked list. */
    if(tmpparamp != NULL) {
      tmpparamp->next = tagp->first;
      tagp->first = tmpparamp;
    }
  }

  return tagp;
}

/**
 * Convert a hexadecimal digit into its corresponding decimal value.
 *
 * @param hex The hexadecimal digit.
 *
 * @return the decimal value.
 */
static int parse_convert_hex(char hex)
{
  if(hex >= '0' && hex <= '9')
    return hex - '0';
  if(hex >= 'a' && hex <= 'f')
    return hex - 'a' + 10;

  return 0;
}

/**
 * Converts a string in the form #rrggbb into a 24-bit colour value.
 * Later, this will also handle named colours.
 *
 * @param colour A string containing the colour code.
 *
 * @return a 24-bit colour value.
 */
u_int32_t parse_convert_colour(char *colour)
{
  u_int32_t real_colour;
  int red, green, blue;

  if(colour[0] != '#' || strlen(colour) != 7)
    return 0xffffff;

  red = parse_convert_hex(tolower(colour[1])) * 16 +
    parse_convert_hex(tolower(colour[2]));
  green = parse_convert_hex(tolower(colour[3])) * 16 +
    parse_convert_hex(tolower(colour[4]));
  blue = parse_convert_hex(tolower(colour[5])) * 16 +
    parse_convert_hex(tolower(colour[6]));

  real_colour = ((red   << 16) |
		 (green <<  8) |
		 (blue  <<  0));

  return real_colour;
}


static char *internal_string = NULL;
static int internal_allocation = 0;
static int is_preformatted = 0;

/**
 * Store one character at the end of the temporary internal string.
 * More than one whitespace characters in a row are ignored. All
 * whitespace characters are converted into space characters.
 *
 * @param c The character to be stored.
 *
 * @return non-zero value if an error occurred.
 */
int parse_string_store_character(char c)
{
  int size;

  /* Convert all whitespace characters into spaces,
   * unless we are in preformating mode. 
   */
  if(!is_preformatted && isspace(c))
    c = ' ';

  /* If we are preformating, newline characters should create what is
   * equivalent to a br-tag.
   */
  if(is_preformatted && c == '\n') {
    struct layout_part *partp;

    parse_string_store_current();
    partp = layout_init_part(LAYOUT_PART_PARAGRAPH);
    if(partp == NULL) {
      return 1;
    }
    layout_add_part(partp);    
  }

  if(internal_string == NULL) {
    internal_string = malloc(1024);
    internal_string[0] = '\0';
    internal_allocation = 1024;
  } else if(strlen(internal_string) == internal_allocation - 1) {
    char *tmp;

    tmp = realloc(internal_string, internal_allocation + 1024);
    if(tmp == NULL)
      return 1;
    internal_string = tmp;
    internal_allocation += 1024;
  }
  
  size = strlen(internal_string);

  /* Skip several whitespaces. */
  if(!is_preformatted && 
     size > 0 && internal_string[size - 1] == ' ' && c == ' ')
    return 0;

  internal_string[size] = c;
  internal_string[size + 1] = '\0';

  return 0;
}

/**
 * Sets the static variable is_preformatted, simply because I do
 * not want to make it global. Modularity, my friend.
 *
 * @member value The new value of is_preformatted.
 */
void parse_string_set_preformatted(int value)
{
  is_preformatted = value;
}

/**
 * Copy the stored string into the provided buffer `buf'. It will 
 * copy the string, or a maximum of `size' characters, in which case
 * the string will not be nul terminated.
 * After the string is copied, the internal string is cleared.
 *
 * @param buf A buffer into which the string goes.
 * @param size Maximum number of characters that are copied.
 *
 * @return non-zero value if an error occurred.
 */
int parse_string_get_stored(char *buf, int size)
{
  if(internal_string == NULL || buf == NULL)
    return 1;

  strncpy(buf, internal_string, size);

  free(internal_string);
  internal_string = NULL;

  return 0;
}

/**
 * Copy the string in buf into the internal string. To be able to
 * do that safely, first discard the old string, as if it were not
 * worth anything at all. 
 *
 * @param buf A buffer containing the string to store.
 *
 * @return non-zero value if an error occurred.
 */
int parse_string_set_stored(char *buf)
{
  int i;

  if(buf == NULL)
    return 1;

  parse_string_discard();
  
  for(i = 0 ; i < strlen(buf) ; i++)
    parse_string_store_character(buf[i]);

  return 0;
}

/**
 * Trims the internal string from whitespace characters at the front and 
 * at the back of the string. If there were whitespace characters at the
 * front, we keep one of them and trim the rest off. The same thing goes
 * for the whitespace characters at the back.
 *
 * @param skip_front_space If non-zero, do not keep any leading space.
 *
 * @return non-zero value if an error occurred.
 */
int parse_string_trim(int skip_front_space)
{
  char *stringp_front, *stringp_back;
  int front_space = 0, back_space = 0;

  if(internal_string == NULL || strlen(internal_string) == 1)
    return 0;

  if(isspace(internal_string[0]))
    front_space = 1;
  if(isspace(internal_string[strlen(internal_string) - 1]))
    back_space = 1;

  /* Find the front. */
  stringp_front = internal_string;
  while(isspace(*stringp_front))
    stringp_front++;
  if(front_space)
    stringp_front--;

  /* Cut the back, but only if there is something left to cut. */
  if(strlen(stringp_front) > 0) {
    stringp_back = internal_string + strlen(internal_string) - 1;
    while(isspace(*stringp_back))
      stringp_back--;
    if(back_space)
      stringp_back++;

    stringp_back[1] = '\0';

    /* Move the trimmed string to the first position in the internal string. */
    memmove(internal_string, stringp_front, strlen(stringp_front) + 1);
  } else {
    internal_string[0] = '\0';
  }

  return 0;
}

/**
 * Return the length of the internal string. 
 *
 * @return the length of the string.
 */
int parse_string_get_length(void)
{
  if(internal_string == NULL)
    return 0;

  return strlen(internal_string);
}

/**
 * Empty, free and nullify the internal string.
 */
void parse_string_discard(void)
{
  if(internal_string != NULL)
    free(internal_string);

  internal_string = NULL;
}

/**
 * Put the currently stored string as a textual layout part.
 * This reads the current text style from the style state stack, and
 * uses the right style for the right text.
 *
 * @return a pointer to the layout part that was just added to the list
 * @return or NULL if there was no string stored.
 */
struct layout_part *parse_string_store_current(void)
{
  char *text;
  int string_length;
  struct layout_part *partp;
  struct layout_text_styles current_style;
  static int last_was_space = 0;
  
  /* Get the current styles of the text to be stored. */
  parse_state_get_current(&current_style, NULL, NULL);

  /* If the page says that this particular piece of text is to be rendered
   * preformatted, then so be it, and we must not modify it by cutting off
   * excess whitespaces.
   */
  if(!current_style.preformatted) {
    parse_string_trim(last_was_space);
  }

  string_length = parse_string_get_length();
  if(string_length > 0) {
    text = malloc(string_length + 1);
    if(text == NULL)
      return NULL;

    parse_string_get_stored(text, string_length + 1);
    partp = layout_init_part(LAYOUT_PART_TEXT);
    if(partp == NULL) {
      free(text);
      return NULL;
    }

    partp->data.text.text = text;
    partp->data.text.style = current_style;
    layout_add_part(partp);

    if(isspace(text[strlen(text) - 1]))
      last_was_space = 1;
    else
      last_was_space = 0;
  } else {
    /* If the string was trimmed away completely, we still need to discard
     * what might be left of it. 
     */
    parse_string_discard();
    return NULL;
  }

  return partp;
}

/**
 * Put the currently stored string as the title for the page. 
 * What is special about this is that it is always placed first in the
 * linked list or parts. 
 *
 * @return non-zero value if an error occurred.
 */
int parse_string_store_title(void)
{
  char *text;
  int string_length;
  struct layout_part *partp;
  
  parse_string_trim(0);
  string_length = parse_string_get_length();
  if(string_length > 0) {
    text = malloc(string_length + 1);
    if(text == NULL)
      return 1;
    parse_string_get_stored(text, string_length + 1);
  } else {
    /* If the string was trimmed away, save NULL in the title text. */
    parse_string_discard();
    text = NULL;
  }

  partp = layout_get_page_information(NULL);

  if(partp == NULL) {
    partp = layout_init_part(LAYOUT_PART_PAGE_INFORMATION);
    if(partp == NULL) {
      if(text)
	free(text);
      return 1;
    }
    layout_add_part(partp);
  }
  
  /* At this point, we will change the real part by changing what is in
   * the pointer partp.
   */
  partp->data.page_information.title = text;

  return 0;
}


/**
 * Prints the currently stored string.
 */
void debug_dump_string(void)
{
  if(internal_string)
    fprintf(stderr, "String: '%s'\n", internal_string);
  else 
    fprintf(stderr, "String: (null)\n");
}

/**
 * Prints all contents of a parse_param struct.
 *
 * @param paramp The parse_param struct to print out.
 */
void debug_dump_param(struct parse_param *paramp)
{
  fprintf(stderr, "Name: %s", paramp->name);

  if(paramp->value != NULL)
    fprintf(stderr, "   Value: %s", paramp->value);

  fprintf(stderr, "\n");
}

/**
 * Prints all contents of a tag struct, including all parameters
 * in the linked list.
 *
 * @param tag The parse_tag struct to print out.
 */
void debug_dump_tag(struct parse_tag *tagp)
{
  struct parse_param *paramp;

  fprintf(stderr, "-- beginning of tag --\n");

  fprintf(stderr, "Name: %s\n", tagp->name);
  fprintf(stderr, "Type: ");
  if(tagp->type == PARSE_TAG_START)
    fprintf(stderr, "Start\n");
  else if(tagp->type == PARSE_TAG_END)
    fprintf(stderr, "End\n");
  else
    fprintf(stderr, "Unknown (This should not happen!)\n");

  /* Dump all parameters in the linked list. */
  paramp = tagp->first;
  if(paramp != NULL)
    fprintf(stderr, "Parameters: \n");
  while(paramp != NULL) {
    debug_dump_param(paramp);
    paramp = paramp->next;
  }

  fprintf(stderr, "-- end of tag --\n");
}

