/**
 * Functions to provide a generic method of getting a stream from 
 * any supported protocol, according to the URL.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "protocol.h"
#include "streams.h"
#include "file.h"
#include "http.h"

/**
 * A pointer to the first element in the linked list containing the 
 * information about the streams. 
 */
static struct protocol_stream *first_stream = NULL;

/* Used to store the current base URL. */
char *base_host = NULL;
char *base_path = NULL;

/**
 *
 */
static int protocol_store_base(char *url)
{
  char *host_text, *path_text, *tmp;

  host_text = (char *)malloc(strlen(url) + 16);
  if(host_text == NULL)
    return -1;
  strcpy(host_text, url);

  if(!strncmp(host_text, "http://", 7)) {
    /* If the URL contains GET method information after a question mark, 
     * we will need to remove that first. We search for the first occurrance
     * of a question mark and remove everything efter that. 
     */
    tmp = strchr(host_text, '?');
    if(tmp != NULL)
      tmp[0] = '\0';

    path_text = index(&host_text[7], '/');
    if(path_text != NULL) {
      path_text[0] = '\0';
      path_text++;
    }
    base_host = realloc(base_host, strlen(host_text) + 1);
    strcpy(base_host, host_text);
  } else if(!strncmp(host_text, "file://", 7)) {
    path_text = host_text + 7;
    base_host = realloc(base_host, 16);
    strcpy(base_host, "file://");
  } else {
    base_host = realloc(base_host, 16);
    strcpy(base_host, "file://");
    path_text = host_text;
  }

  /* Try to intelligently figure out if we need to add an extra slash
   * at the end of the line, just to be compatible with people who have
   * not learned how to write a proper URL yet.
   */
  if(path_text[strlen(path_text) - 1] != '/') {
    tmp = rindex(path_text, '/');
    if(tmp == NULL)
      tmp = path_text;
    if(index(tmp, '.') != NULL) {
      tmp[0] = '\0';
    }
  } else {
    path_text[strlen(path_text) - 1] = '\0';
  }  
  base_path = realloc(base_path, strlen(path_text) + 1);
  strcpy(base_path, path_text);

  return 0;
}

/**
 * Take a URL and give it an awful treatment, to squeeze the size to its
 * very limits. The lower limits. More detailed, this will remove double,
 * or more, slashes, slash dot slash and slash dot dot together with the 
 * previous directory. We do not remove slash dot at the end of a URL,
 * because Roxen web server uses this in a feature to list a directory
 * where index.html, or similar, normally would have been loaded.
 * This function expects an absolute URL.
 *
 * @param url The uncompressed URL.
 *
 * @return non-zero value if an error occurred.
 */
static int protocol_compress_url(char *url)
{
  char *dir_start, *tmp, *tmp2;

  if(!strncmp(url, "http://", 7)) {
    dir_start = index(&url[7], '/');
    if(dir_start == NULL)
      return 0;
  } else if(!strncmp(url, "file://", 7)) {
    dir_start = &url[7];
  } else {
    return 1;
  }

  /* Start with removing all multiple slashes. */
  while((tmp = strstr(dir_start, "//")) != NULL) {
    memmove(tmp, &tmp[1], strlen(tmp));
  }

  /* Then remove slash dot slash. */
  while((tmp = strstr(dir_start, "/./")) != NULL) {
    memmove(tmp, &tmp[2], strlen(tmp) - 1);
  }

  /* And almost last, but almost not least, remove slash dot dot slash
   * and previous directory.  
   */
  while((tmp = strstr(dir_start, "/../")) != NULL) {
    tmp2 = tmp;
    tmp2--;
    while(tmp2 >= dir_start && *tmp2 != '/') {
      tmp2--;
    }
    if(tmp2 < dir_start)
      memmove(tmp, &tmp[3], strlen(tmp) - 2);
    else
      memmove(tmp2, &tmp[3], strlen(tmp) - 2);
  }  

  /* Last case, which is a special treatment of slash dot dot, at the
   * end of a URL.
   */
  tmp = strstr(dir_start, "/..");
  if(tmp != NULL && tmp[3] == '\0') {
    tmp2 = tmp;
    tmp2--;
    while(tmp2 >= dir_start && *tmp2 != '/') {
      tmp2--;
    }
    if(tmp2 < dir_start)
      memmove(tmp, &tmp[3], strlen(tmp) - 2);
    else
      memmove(tmp2, &tmp[3], strlen(tmp) - 2);    
  }

  return 0;
}

/**
 * This takes a URL, absolute or relative, and tries to create an 
 * absolute URL out of it, and the stored base URL. 
 *
 * Comments: This does not yet take care of ../ and ./ as directories
 *           but nicely just adds them to the final URL. This will
 *           be fixed.
 *
 * @param url The absolute or relative URL to absolutify.
 * @param base_url If not NULL, this is used as the new base URL.
 *
 * @return the new, absolutely absolute URL.
 */
static char *protocol_make_absolute(char *url, char *base_url)
{
  char *new_url;
  int alloc_size;
  char cwd[2048];

  /* First of all, we check if an unsupported protocol is being used.
   * If it is, we do not proceed. 
   * This is a bad method, since we cannot possible list all protocols
   * we will not support. The problem is that if a known protocol is
   * not specified, we want to try to access it as a file.
   */
  if(!strncmp(url, "ftp://", 6) ||
     !strncmp(url, "mailto:", 7) ||
     !strncmp(url, "telnet://", 9)) {
    fprintf(stderr, "Unsupported protocol used for '%s'.\n", url);
    return NULL;
  }

  /* If a new base URL was provided, we set this as the current base URL. */
  if(base_url)
    protocol_store_base(base_url);

  if(!strncmp(url, "http://", 7) ||
     !strncmp(url, "file://", 7)) {
    /* New absolute URL. */
    new_url = (char *)malloc(strlen(url) + 1);
    if(new_url == NULL)
      return NULL;
    strcpy(new_url, url);

  } else if(url[0] == '/') {
    /* New path only. */
    if(base_host)
      alloc_size = strlen(base_host) + strlen(url) + 1;
    else
      alloc_size = 7 + strlen(url) + 1;

    new_url = (char *)malloc(alloc_size);
    if(new_url == NULL)
      return NULL;

    if(base_host)
      strcpy(new_url, base_host);
    else
      strcpy(new_url, "file://");
    strcat(new_url, url);

  } else {
    /* Relative path. */
    alloc_size = strlen(url) + 4;
    if(base_host)
      alloc_size += strlen(base_host);
    else
      alloc_size += 7;
    if(base_path) {
      alloc_size += strlen(base_path);
    } else {
      if(base_host == NULL) {
	if(getcwd(cwd, 2047) == NULL) {
	  return NULL;
	}
	alloc_size += strlen(cwd);
      }
    }
  
    new_url = (char *)malloc(alloc_size);
    if(new_url == NULL)
      return NULL;

    if(base_host) {
      strcpy(new_url, base_host);
      strcat(new_url, "/");
    } else {
      strcpy(new_url, "file://");
    }  

    if(base_path) {
      strcat(new_url, base_path);
    } else {
      if(base_host == NULL) {
	strcat(new_url, cwd);
      }
    }
    strcat(new_url, "/");
    strcat(new_url, url);
  }

  protocol_compress_url(new_url);

  return new_url;
}

/**
 * Takes a good look at the specified URL, looks deeply into its darkest
 * attributes and corners of significancy. After that, it tries its best
 * to decide on which protocol to use to achieve a stream from the URL.
 *
 * @param url The URL to open a stream from.
 * @param referer The URL we got from to get here, or NULL if jumping here.
 * @param base_url The base URL to be used to create an absolute URL from 
 * @param base_url the given URL, or NULL to use the locally stored base.
 *
 * @return a file descriptor for the stream, or a negative value if an
 * @return error occurred.
 */
int protocol_open(char *url, char *referer, char *base_url)
{
  int fd;
  char *new_url;
  enum protocol_type protocol;
  struct protocol_stream *new_stream;
  struct protocol_http_headers *headers;

  headers = (struct protocol_http_headers *)
    malloc(sizeof(struct protocol_http_headers));
  if(headers == NULL) {
    return -1;
  }

  /* Fill in the default values for the HTTP headers struct. */
  headers->return_code = 200;
  headers->return_message = NULL;
  headers->content_type_major = NULL;
  headers->content_type_minor = NULL;
  headers->content_length = -1;
  headers->server = NULL;
  headers->real_url = NULL;

  if(referer)
    protocol_store_base(referer);

  /* Get the absolute equivalence to the specified URL. */
  new_url = protocol_make_absolute(url, base_url);

  protocol = PROTOCOL_UNKNOWN;
  if(new_url) {
    /* Find the correct protocol to use. */
    if(!strncmp(new_url, "http://", 7)) {
      fd = protocol_http_open(&new_url[7], referer, headers);
      protocol = PROTOCOL_HTTP;
    } else if(!strncmp(new_url, "file://", 7)) {
      fd = protocol_file_open(&new_url[7], headers);
      protocol = PROTOCOL_FILE;
    } else {
      fprintf(stderr, "Unsupported protocol used for '%s'\n", new_url);
      fd = -1;
    }
  } else {
    fd = -1;
  }

  /* Unless there was an unexpected error, we put the new file descriptor
   * and information about it onto the linked list. Its much easier to put
   * it at the top of the list, so we do that.
   */
  if(fd < 0) {
    if(new_url)
      free(new_url);
    return fd;
  }

  /* This is not really right. */
  if(headers->return_code != 200) {
    free(new_url);
    protocol_close(fd);
    return -1;
  }

  new_stream = (struct protocol_stream *)
    malloc(sizeof(struct protocol_stream));
  if(new_stream == NULL) {
    if(headers)
      free(headers);
    protocol_close(fd);
    return -1;
  }
  new_stream->fd = fd;
  new_stream->protocol = protocol;
  new_stream->headers = headers;
  new_stream->next = first_stream;
  first_stream = new_stream;

  /* If we open a HTML page, we should set the base URL strings to something
   * apropriate.
   */
  if(!strcmp(headers->content_type_major, "text") &&
     !strcmp(headers->content_type_minor, "html"))
    protocol_store_base(headers->real_url);

  free(new_url);

#if 0
  if(new_stream->headers) {
    fprintf(stderr, "return code: %d\n", 
	    new_stream->headers->return_code);
    fprintf(stderr, "return message: %s\n", 
	    new_stream->headers->return_message);
    fprintf(stderr, "content_type: %s/%s\n", 
	    new_stream->headers->content_type_major, 
	    new_stream->headers->content_type_minor);
    fprintf(stderr, "content_length: %d\n", 
	    new_stream->headers->content_length);
    fprintf(stderr, "server: %s\n", 
	    new_stream->headers->server);
    fprintf(stderr, "real_url: %s\n", 
	    new_stream->headers->real_url);
  }
#endif

  return fd;
}

/**
 * Closes a stream that was opened by protocol_stream_open(). It has 
 * to have been opened by that function, or it will not work. This
 * will free the memory that was allocated for the HTTP headers.
 *
 * @param fd The file descriptor of the stream to close.
 *
 * @return non-zero value if an error occurred.
 */
int protocol_close(int fd)
{
  struct protocol_stream *streamp, *previous_stream;

  /* Find the stream associated with the specified file descriptor. */
  previous_stream = NULL;
  streamp = first_stream;
  while(streamp && streamp->fd != fd) {
    previous_stream = streamp;
    streamp = streamp->next;
  }

  /* Maybe we cannot find the file descriptor in the list. What can we do
   * but to exit and tell the caller that there was an error. A thoughtful
   * alternative would be to use close() to close the unknown file
   * descriptor, but that is not really recommended.
   */
  if(streamp == NULL)
    return 1;

  /* Close the stream, and remove it from the linked list. */
  switch(streamp->protocol) {
  case PROTOCOL_HTTP:
    protocol_http_close(fd);
    break;

  case PROTOCOL_FILE:
    protocol_file_close(fd);
    break;

  default:
    fprintf(stderr, 
	    "%s: Unsupported protocol detected. Sorry, cannot close.\n",
	    __FUNCTION__);
  }

  if(previous_stream)
    previous_stream->next = streamp->next;
  else
    first_stream = streamp->next;
  
  if(streamp->headers) {
    if(streamp->headers->return_message)
      free(streamp->headers->return_message);
    if(streamp->headers->content_type_major)
      free(streamp->headers->content_type_major);
    if(streamp->headers->content_type_minor)
      free(streamp->headers->content_type_minor);
    if(streamp->headers->server)
      free(streamp->headers->server);
    if(streamp->headers->real_url)
      free(streamp->headers->real_url);

    free(streamp->headers);
  }
  free(streamp);

  return 0;
}

/**
 * Get the HTTP headers associated with a particular stream.
 *
 * @param fd The file descriptor of the stream.
 *
 * @return a pointer to the struct holding the header information, or NULL
 * @return if the stream did not have any HTTP headers, or if the stream
 * @return could not be found for the given file descriptor.
 */
struct protocol_http_headers *protocol_get_headers(int fd)
{
  struct protocol_stream *streamp;

  /* Find the stream associated with the specified file descriptor. */
  streamp = first_stream;
  while(streamp && streamp->fd != fd) {
    streamp = streamp->next;
  }
  if(streamp == NULL)
    return NULL;

  return streamp->headers;
}
