/**
 * Functions to open a stream from a server using the HTTP protocol.
 * This currently only supports HTTP/1.0 but that is enough for most
 * things.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

#include "ui.h"
#include "http.h"

/**
 * Open a TCP connection to the other host, on a specific port number. 
 *
 * @param host A string containing the name of the host. Either a DNS
 * @param host name, or an IPv4 number is allowed.
 * @param port The number of the port to connect to.
 *
 * @return a file descriptor for the opened stream, or a negative value
 * @return representing an error.
 */
static int protocol_http_open_connection(char *host, short port)
{
  struct hostent *hent;
  struct protoent *proto;
  struct sockaddr_in saddr;
  int sock;

  proto = getprotobyname("tcp");
  if(proto == NULL) {
    return -1;
  }

  sock = socket(AF_INET, SOCK_STREAM, proto->p_proto);
  if(sock < 0) {
    return -1;
  }

  hent = gethostbyname(host);
  if(hent == NULL) {
    close(sock);
    return -1;
  }
  memcpy(&saddr.sin_addr, hent->h_addr_list[0], sizeof(struct in_addr));
  saddr.sin_family = AF_INET;
  saddr.sin_port = htons(port);
  
  if(connect(sock, (struct sockaddr *)&saddr, sizeof(saddr)) < 0) {
    close(sock);
    return -1;
  }

  return sock;
}

/**
 * Opens an HTTP stream from a web server. 
 *
 * @param url The name of the URL to open.
 * @param referer The URL we were at when entering this new URL.
 * @param headers A pointer to a struct meant to contain information gathered
 * @param headers from the HTTP headers in the response from the server. If
 * @param headers this is set to NULL, do not even try to store any headers.
 *
 * @return the file descriptor for the http stream or a negative value
 * @return if an error occurred.
 */
int protocol_http_open(char *url, char *referer, 
		       struct protocol_http_headers *headers)
{
  int fd;
  char *host_text, *port_text, *file_text, *request, *referer_text;
  short port;
  char c, header_row[1024], *tmp;
  int i, error, alloc_size;

  /* First of all, copy the URL into our variable, so that we do not disturb
   * the sense of order and style that the previous function was trying to
   * induce upon the greater good.
   */
  host_text = (char *)malloc(strlen(url) + 16);
  if(host_text == NULL)
    return -1;
  strcpy(host_text, url);

  /* We pick out the host, port number and file/directory to retreive the
   * data from. The file/directory will be stripped of its leading slash
   * in this brutal process, so we will have to add that later, when making
   * the actual HTTP request.
   */
  port_text = index(host_text, ':');
  file_text = index(host_text, '/');
  if(file_text != NULL) {
    file_text[0] = '\0';
    file_text++;
  } else {
    file_text = "";
  }

  /* Convert the port number from text to a real number. If the port was not
   * specified in the URL, we use the default HTTP port, which is number 80.
   */
  if(port_text != NULL) {
    port_text[0] = '\0';
    port_text++;
    port = atoi(port_text);
  } else {
    port = 80;
  }

  fd = protocol_http_open_connection(host_text, port);

  /* Create header text for a referer header, if referer is given. */
  if(referer != NULL) {
    referer_text = (char *)malloc(strlen(referer) + 16);
    if(referer_text == NULL) {
      free(host_text);
      return -1;
    }
    sprintf(referer_text, "Referer: %s\r\n", referer);
  } else {
    referer_text = "";
  }

  /* Create the HTTP request to retreive an object from the server. */
  request = (char *)malloc(16384);
  sprintf(request, 
	  "GET /%s HTTP/1.0\r\n"
	  "User-Agent: Zen/%s %s/%s\r\n"
	  "Host: %s\r\n"
	  "Accept: */*\r\n"
	  "%s"
	  "\r\n", 
	  file_text, 
	  VERSION, user_interface.ui_name, user_interface.ui_version_text,
	  host_text,
	  referer_text);
  write(fd, request, strlen(request));

  /* Take care of the response from the server. 
   * Fill in the header struct with appropriate values.
   * This also have to take into account stupidly implemented HTTP servers
   * which send only LF instead of CRLF as the RFC clearly specifies.
   */
  error = 0;
  /* First read the response code. */
  headers->return_code = 404;
  if(read(fd, header_row, 9) > 0) {
    header_row[9] = '\0';
    if(!strncmp(header_row, "HTTP/", 5) && read(fd, header_row, 4) > 0) {
      headers->return_code = atoi(header_row);
      i = 0;
      while(read(fd, &c, 1) > 0 && c != '\r' && c != '\n') {
	if(i < 1023)
	  header_row[i++] = c;
      }
      if(c == '\r')
	read(fd, &c, 1);
      if(i > 0) {
	header_row[i] = '\0';
	tmp = (char *)malloc(strlen(header_row) + 1);
	if(tmp == NULL)
	  error = 1;
	strcpy(tmp, header_row);
	headers->return_message = tmp;      
      }
    }
  }
 
  /* If it could not find the page, try adding an ending slash, to get it as
   * a directory. I will implement this in a better way later.
   */
  if(headers->return_code != 200) {
    /* Close the old stream. */
    close(fd);

    /* Open it again. */
    fd = protocol_http_open_connection(host_text, port);

    /* Create the HTTP request to retreive an object from the server. */
    sprintf(request, 
	    "GET /%s/ HTTP/1.0\r\n"
	    "User-Agent: Zen/%s %s/%s\r\n"
	    "Host: %s\r\n"
	    "Accept: */*\r\n"
	    "%s"
	    "\r\n", 
	    file_text, 
	    VERSION, user_interface.ui_name, user_interface.ui_version_text,
	    host_text,
	    referer_text);
    write(fd, request, strlen(request));

    /* Take care of the response from the server. 
     * Fill in the header struct with appropriate values.
     */
    error = 0;  
    /* First read the response code. */
    headers->return_code = 404;
    if(read(fd, header_row, 9) > 0) {
      header_row[9] = '\0';
      if(!strncmp(header_row, "HTTP/", 5) && read(fd, header_row, 4) > 0) {
	headers->return_code = atoi(header_row);
	i = 0;
	while(read(fd, &c, 1) > 0 && c != '\r' && c != '\n') {
	  if(i < 1023)
	    header_row[i++] = c;
	}
	if(c == '\r')
	  read(fd, &c, 1);
	if(i > 0) {
	  header_row[i] = '\0';
	  tmp = (char *)malloc(strlen(header_row) + 1);
	  if(tmp == NULL)
	    error = 1;
	  strcpy(tmp, header_row);
	  headers->return_message = tmp;      
	}
      }
    }
  }

  free(request);

  /* Store the complete URL used for reading. */
  alloc_size = 7 + strlen(host_text) + strlen(file_text) + 16;
  if(port_text)
    alloc_size += strlen(port_text);
  headers->real_url = (char *)malloc(alloc_size);
  if(headers->real_url == NULL) {
    error = 1;
  } else {
    if(port_text)
      sprintf(headers->real_url, "http://%s:%s/%s", 
	      host_text, port_text, file_text);
    else
      sprintf(headers->real_url, "http://%s/%s", 
	      host_text, file_text);
  }

  /* Read all headers and save the valuable ones. 
   * Again, make sure unproper HTTP servers also gets treated.
   */
  i = 0;
  while(!error && read(fd, &c, 1) > 0) {
    if(c == '\r' || c == '\n') {
      char *field, *value;

      /* If we come to a CR, always expect an LF. */
      if(c == '\r')
	read(fd, &c, 1);

      /* If no characters have been read since previous row, it means the
       * actual stream data is coming next, right after another line feed.
       */
      if(i == 0 && c == '\n')
	break;

      header_row[i] = '\0';
      i = 0;
      field = header_row;
      value = strstr(header_row, ": ");
      if(value == NULL)
	continue;
      value[0] = '\0';
      value += 2;

      /*       fprintf(stderr, "field: '%s', value: '%s'\n", field, value); */

      /* Figure out where to put the data. */
      if(!strcasecmp(field, "content-type")) {
	char *divider;

	tmp = (char *)malloc(strlen(value) + 1);
	if(tmp == NULL) {
	  error = 1;
	  continue;
	}
	strcpy(tmp, value);
	divider = index(tmp, '/');
	if(divider == NULL)
	  continue;
	divider[0] = '\0';
	headers->content_type_major = tmp;

	tmp = (char *)malloc(strlen(value) + 1);
	if(tmp == NULL) {
	  error = 1;
	  continue;
	}
	strcpy(tmp, &divider[1]);
	headers->content_type_minor = tmp;
      } else if(!strcasecmp(field, "server")) {
	tmp = (char *)malloc(strlen(value) + 1);
	if(tmp == NULL) {
	  error = 1;
	  continue;
	}
	strcpy(tmp, value);
	headers->server = tmp;
      } else if(!strcasecmp(field, "content-length")) {
	headers->content_length = atol(value);
      }
    } else {
      if(c != '\n' && i < 1023)
	header_row[i++] = c;
    }
  }

  free(host_text);
  if(referer_text != NULL && strlen(referer_text) > 0)
    free(referer_text);

  return fd;
}

/**
 * Close an HTTP stream.
 *
 * @param fd The file descriptor associated with the stream to close.
 */
int protocol_http_close(int fd)
{
  close(fd);

  return 0;
}
