/**
 * Functions to render parts of a page in the oFBis user interface.
 *
 * @author Tomas Berndtsson <tomas@nocrew.org>
 */

/*
 * Copyright (C) 1999, Tomas Berndtsson <tomas@nocrew.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ofbis.h>

#include "zen_ui.h"
#include "info.h"

/* The linked list of links. */
struct ofbis_link *links = NULL;

/**
 * Adds a link to the linked list of links.
 *
 * @param partp A pointer to a layout part representing the object 
 * @param partp which is shown as a link.
 * @param linkp A pointer to the layout part containing the link information.
 */
static void add_link(struct layout_part *partp, struct layout_part *linkp)
{
  struct ofbis_link *new, *lastp;
  
  new = (struct ofbis_link *)malloc(sizeof(struct ofbis_link));
  if(new == NULL)
    return;

  new->link = partp->geometry;
  new->part = linkp;
  new->visual_part = partp;

  if(links == NULL) {
    new->next = NULL;
    new->previous = NULL;
    links = new;
  } else {
    /* Find the last element in the list of links. */
    lastp = links;
    while(lastp->next)
      lastp = lastp->next;

    /* Put the link last in the list. */
    new->next = NULL;
    new->previous = lastp;
    lastp->next = new;
  }
}

/**
 * Delete all previous links stored on the linked list of links.
 */
static void delete_all_links(void)
{
  struct ofbis_link *linkp, *tmp;

  linkp = links;
  while(linkp) {
    tmp = linkp;
    linkp = linkp->next;
    free(tmp);
  }

  links = NULL;
}

/**
 * Render the page by calling the appropriate rendering functions.
 *
 * @param parts A pointer to the first layout part in the list to be rendered.
 * @param geometry The offset and size describing what to render.
 *
 * @return non-zero value if an error occurred.
 */
int ofbis_render(struct layout_part *parts, struct layout_rectangle geometry)
{
  struct ofbis_information *info = 
    (struct ofbis_information *)ofbis_ui->ui_specific;
  struct layout_part *partp;
  struct layout_rectangle full_geometry, control_geometry;
  static struct layout_part *linkp = NULL;

  partp = parts;

  /* Make sure we only remove links from the list when this function is
   * called for the first time, for this page. Do some other things then
   * too.
   */
  if(partp && partp->type == LAYOUT_PART_PAGE_INFORMATION) {
    ofbis_move_mouse(0, 0, 1);
    linkp = NULL;
    delete_all_links();
  }

  while(partp) {
    /* Check that the part will fit within the requested display. This
     * does not yet perform any clipping of the parts, so whole parts will
     * be left out.
     */
    if(partp->type != LAYOUT_PART_PAGE_INFORMATION && 
       partp->type != LAYOUT_PART_LINK && 
       (partp->geometry.x_position < geometry.x_position ||
	(partp->geometry.x_position + partp->geometry.width > 
	 geometry.x_position + geometry.width) ||
	partp->geometry.y_position < geometry.y_position ||
	(partp->geometry.y_position + partp->geometry.height > 
	 geometry.y_position + geometry.height))) {

      partp = partp->next;

      continue;
    }

    /* Only the supported parts are represented here, all the others are
     * covered by the default statement, which does not do anything.
     */
    switch(partp->type) {
    case LAYOUT_PART_PAGE_INFORMATION:
      info->default_background_colour = 
	partp->data.page_information.background_colour;
      info->default_text_colour = 
	partp->data.page_information.text_colour;
      info->default_link_colour = 
	partp->data.page_information.link_colour;
      info->default_active_link_colour = 
	partp->data.page_information.active_link_colour;
      info->default_visited_link_colour = 
	partp->data.page_information.visited_link_colour;

      /* We cannot simply set the background colour to a new colour, since 
       * the images will be mapped to the original palette. Instead, we should
       * fill the screen with the colour. This is possible here, since we know
       * that this part is always the first one in the list.
       */
      full_geometry.x_position = info->x_position;
      full_geometry.y_position = info->y_position;
      full_geometry.width = geometry.width;
      full_geometry.height = geometry.height;
      ofbis_fill_area(full_geometry, info->default_background_colour);
      control_geometry.x_position = 0;
      control_geometry.y_position = 0;
      control_geometry.width = ofbis_ui->ui_display.width;
      control_geometry.height = 16;
      ofbis_fill_area(control_geometry, info->control_field_colour);
      ofbis_set_title(partp->data.page_information.title);
      ofbis_set_url(partp->data.page_information.url);

      ofbis_render(partp->child, geometry);
      
      /* We know that when we return from the above rendering the whole page is
       * rendered. Therefore, it is simplest to return from this function here.
       */
      return 0;
      break;

    case LAYOUT_PART_TEXT:
      if(linkp)
	add_link(partp, linkp);
      ofbis_draw_text(partp, geometry);
      break;

    case LAYOUT_PART_LINK:
      linkp = partp;
      ofbis_render(partp->child, geometry);
      break;

    case LAYOUT_PART_GRAPHICS:
      if(linkp) {
	add_link(partp, linkp);
	if(partp->data.graphics.border == 0)
	  partp->data.graphics.border = 1;
	partp->data.graphics.border_colour = info->default_link_colour;
      }
      ofbis_draw_image(partp, geometry);
      break;

    case LAYOUT_PART_LINE:
      ofbis_draw_line(partp, geometry);
      break;

    default:
      break;
    }

    partp = partp->next;
  }

  linkp = NULL;

  return 0;
}

/**
 * Render the given part in active link colour, and restore the previously
 * active link to its original colour.
 *
 * @param linkp A pointer to the link which is going to become the new 
 * @param linkp active link. If NULL, we only reinitialize the previously
 * @param linkp saved active link as NULL, to start over.
 * @param geometry The geometry of the page, so that the rendering functions
 * @param geometry will know where to draw the part. 
 */
void ofbis_activate_link(struct ofbis_link *linkp,
			 struct layout_rectangle geometry)
{
  struct ofbis_information *info = 
    (struct ofbis_information *)ofbis_ui->ui_specific;
  static struct layout_part *previously_active = NULL;
  static u_int32_t previous_colour = 0;
  struct layout_part *partp;

  if(linkp)
    partp = linkp->visual_part;
  else
    partp = NULL;

  /* If the previously active part is the same as the one we are trying to
   * activate now, we do not need to do anything, and therefore we will save
   * time and money by simply returning from the function.
   */
  if(previously_active == partp)
    return;

  /* Restoring the old part. */
  if(previously_active) {
    switch(previously_active->type) {
    case LAYOUT_PART_TEXT:
      previously_active->data.text.style.colour = previous_colour;
      ofbis_render(previously_active, geometry);      
      break;
      
    case LAYOUT_PART_GRAPHICS:
      previously_active->data.graphics.border_colour = previous_colour;
      ofbis_render(previously_active, geometry);
      break;
      
    default:
      fprintf(stderr, "If you see this, the programmer has made a mistake.\n");
    }
  }

  if(linkp == NULL) {
    previously_active = NULL;
    ofbis_set_status_text(NULL);
    return;
  }

  /* Make the current part a new colour. */
  switch(partp->type) {
  case LAYOUT_PART_TEXT:
    previous_colour = partp->data.text.style.colour;
    partp->data.text.style.colour = info->default_active_link_colour;
    ofbis_render(partp, geometry);
    break;

  case LAYOUT_PART_GRAPHICS:
    previous_colour = partp->data.graphics.border_colour;
    partp->data.graphics.border_colour = info->default_active_link_colour;
    ofbis_render(partp, geometry);
    break;

  default:
    fprintf(stderr, "If you see this, the programmer has made a mistake.\n");
  }

  ofbis_set_status_text(linkp->part->data.link.href);
  
  previously_active = partp;
}
